# Copyright (c) 2021 Shuhei Nitta. All rights reserved.
__version__ = "0.0.0"

import setuptools.command.build_ext

from .extension import GoExtension
from .compiler import GoPyCompiler
from .build_ext import build_ext as _build_ext
from . import exceptions

__all__ = [
    "GoExtension",
    "GoPyCompiler",
    "exceptions",
]

setattr(setuptools.command.build_ext, "build_ext", _build_ext)



# Copyright (c) 2021 Shuhei Nitta. All rights reserved.

from setuptools import extension
import typing as t


class GoExtension(extension.Extension):  # type: ignore
    """
    Extension written in Go language.

    Examples
    --------
    >>> from setuptools import setup
    >>> from go_extension import GoExtension

    >>> ext = GoExtension(
    ...     name='hello_go',
    ...     packages=['github.com/huisint/go-extension-python/hello']
    ... )

    >>> setup(ext_modules=[ext])
    """

    _original_name: str
    _packages: list[str]
    _compiled_name: str = "go"
    name: str
    sources: list[str]
    depends: list[str]

    def __init__(
        self,
        name: str,
        packages: t.Sequence[str],
        sources: t.Optional[list[str]] = None,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> None:
        """
        Parameters
        ----------
        name : str
            Package name. Similar to that of setuptools.extension.Exention.
        packages : Sequence[str]
            Sequence of Go package.
            If your project root is also go package root named `github.com/foo/bar`
            (which means 'go.mod' file exists in your progect root),
            and you have go pacakge named 'go_pkg' in your project root,
            set `packages=['github.com/foo/bar/go_pkg']`.
        sources : list[str] | None
            Source files of GoExtension (optional).
            This is used just to judge whether sources have been updated.

        *args, **kwargs : Any
            The same parameters as setuptools.extension.Extension.
        """
        super().__init__(
            f"{name}._{self._compiled_name}",
            sources or [],
            *args,
            **kwargs,  # the same name as .so file gopy creates
        )
        self._original_name = str(name)
        self._packages = [str(pkg) for pkg in packages]

    @property
    def packages(self) -> list[str]:
        """List of Go package."""
        return self._packages

    @property
    def original_name(self) -> str:
        return self._original_name



# Copyright (c) 2021 Shuhei Nitta. All rights reserved.

import os
import sys
import shutil
from pip._internal import metadata
from pip._vendor.packaging import utils
from setuptools._distutils import spawn, log

from go_extension import extension, exceptions


class GoPyCompiler:
    """
    Wrapper class of command `gopy`

    See also
    --------
    https://github.com/go-python/gopy

    """

    go_command: str = "go"
    verbose: bool = False
    dry_run: bool = False
    force: bool = False
    inplace: bool = False
    env: dict[str, str] = dict()

    def __init__(
        self,
        go_command: str = "go",
        verbose: bool = False,
        dry_run: bool = False,
        force: bool = False,
        inplace: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        go_command : str
            Path to go executable command. used as `go mod`, `go get`, etc...

        other parameters : bool
            See distutils.ccompiler.CCompiler.__init__().

        """
        self.go_command = go_command
        self.verbose = verbose
        self.dry_run = dry_run
        self.force = force
        self.inplace = inplace
        self.env = os.environ.copy()
        self.env["LD_LIBRATY_PATH"] = (self.env.get("LD_LIBRATY_PATH", "") + ":" + os.curdir).strip(":")

    def install_build_tools(self) -> None:
        """
        Install gopy, goimports, and pybindgen.

        Raises
        ------
        go_extension.exceptions.GoNotFoundError
            If `go` command not found.
        """
        if not shutil.which(self.go_command):
            # ToDo: should check go version
            raise exceptions.GoNotFoundError("Go1.16 or above is required to build this extension.")

        if not _gopy_is_installed():
            log.info("Installing gopy...")
            cmd = [
                self.go_command,
                "install",
                "github.com/go-python/gopy@latest"
            ]
            self.spawn(cmd)

        if not _goimports_is_installed():
            log.info("Installing goimports...")
            cmd = [
                self.go_command,
                "install",
                "golang.org/x/tools/cmd/goimports@latest",
            ]
            self.spawn(cmd)

        if not _pybindgen_is_installed():
            pybindgen = "pybindgen"
            log.info(f"Installing {pybindgen}...")
            cmd = [sys.executable, "-m", "pip", "install", "-Uq", pybindgen]
            self.spawn(cmd)

    def generate(self, ext: extension.GoExtension, output: str | os.PathLike[str]) -> None:
        """
        Generate (C)Python language bindings for Go.

        Parameters
        ----------
        ext : go_extension.extension.GoExtension
            Go extension.

        output : str | os.PathLike[str]
            Output directory for bindings.

        See also
        --------
        https://github.com/go-python/gopy

        """
        self.install_build_tools()
        cmd: list[str] = [
            "gopy",
            "gen",
            f"-name={ext._compiled_name}",
            "-no-make",
            "-rename",
            f"-vm={sys.executable}",
            f"-output={output}",
        ] + ext.packages
        self.spawn(cmd)

    def build(self, ext: extension.GoExtension, output: str | os.PathLike[str]) -> None:
        """
        Generate and compile (C)Python language bindings for Go.

        Parameters
        ----------
        ext : go_extension.extension.GoExtension
            Go extension.

        output : str | os.PathLike[str]
            Output directory for bindings.

        See also
        --------
        https://github.com/go-python/gopy
        """
        self.install_build_tools()
        cmd: list[str] = [
            "gopy",
            "build",
            f"-name={ext._compiled_name}",
            "-no-make",
            "-rename",
            f"-vm={sys.executable}",
            f"-output={output}",
        ] + ext.packages
        self.spawn(cmd)

    def spawn(self, cmd: list[str]) -> None:
        """
        Run another program, specified as a command list 'cmd', in a new process.

        Parameters
        ----------
        cmd : list[str]
            Argument list for the new process.
        """
        spawn.spawn(cmd, verbose=self.verbose, dry_run=self.dry_run, env=self.env)


def _gopy_is_installed() -> bool:
    return bool(shutil.which("gopy"))


def _goimports_is_installed() -> bool:
    return bool(shutil.which("goimports"))


def _pybindgen_is_installed() -> bool:
    env = metadata.get_default_environment()
    pip_installed = {dist.canonical_name: dist for dist in env._iter_distributions()}
    return bool(utils.canonicalize_name("pybindgen") in pip_installed)



# Copyright (c) 2021 Shuhei Nitta. All rights reserved.

from setuptools.command import build_ext as _build_ext
from setuptools.extension import Extension
from setuptools import dist, dep_util
from setuptools._distutils import log
from pathlib import Path

from go_extension import extension, compiler


class build_ext(_build_ext.build_ext):  # type: ignore
    """
    Command `build_ext` able to build GoExtension.

    If you change `go` command, inherit this class and pass it to setuptools.setup() like this.

    >>> from setuptools import setup
    >>> from go_extension import build_ext
    >>> class build_ext(build_ext.build_ext):
    ...     go_command = 'path/to/go(executable)'
    >>> setup(cmdclass={'build_ext': build_ext})
    """

    go_command: str = "go"
    gopycompiler: compiler.GoPyCompiler

    def __init__(self, distr: dist.Distribution) -> None:
        super().__init__(distr)
        self.gopy_compiler = compiler.GoPyCompiler()

    def run(self) -> None:
        self.gopy_compiler = compiler.GoPyCompiler(
            go_command=self.go_command,
            verbose=bool(getattr(self, "verbose", False)),
            dry_run=bool(getattr(self, "dry_run", False)),
            force=bool(getattr(self, "force", False)),
            inplace=bool(getattr(self, "inplace", False)),
        )
        super().run()

    def build_extension(self, ext: Extension) -> None:
        if isinstance(ext, extension.GoExtension):
            self.build_go(ext)
        else:
            super().build_extension(ext)

    def build_go(self, ext: extension.GoExtension) -> None:
        """
        Build GoExtension.

        Parameters
        ----------
        ext : go_extension.extension.GoExtension
            Go extension.
        """
        assert isinstance(ext, extension.GoExtension)
        if self.should_skip_ext(ext):
            log.debug("skipping '%s' extension (up-to-date)", ext.original_name)
            return
        else:
            log.info("building '%s' extension", ext.original_name)

        if self.gopy_compiler.inplace:
            # reference: setuptools.command.build_ext.build_ext.copy_extensions_to_source()
            build_py = self.get_finalized_command("build_py")
            fullname = str(self.get_ext_fullname(ext.name))
            modpath = fullname.split(".")
            package = ".".join(modpath[:-1])
            package_dir = build_py.get_package_dir(package)
            # end
            self.gopy_compiler.generate(ext, output=Path(package_dir))
        ext_path = self.get_ext_fullpath(ext.name)
        self.gopy_compiler.build(ext, output=Path(ext_path).parent)

    def should_skip_ext(self, ext: Extension) -> bool:
        """
        Return true if the extension should be skipped.

        Parameters
        ----------
        ext : setuptools.extension.Extension

        Returns
        -------
        bool
        """
        ext_path = self.get_ext_fullpath(ext.name)
        depends = ext.sources + ext.depends
        return not (self.force or dep_util.newer_group(depends, ext_path, "newer"))
